﻿/*
      Copyright (C) 2009 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using BEPUphysics.Constraints;
using Microsoft.Xna.Framework;
using BEPUphysics.Entities;
using BEPUphysics;

namespace BEPUphysicsDemos
{
    class SingleBodyLinearMotor : SingleBodyConstraint
    {
        #region Tweaking Variables

        /// <summary>
        /// Location to which this constraint will try to move the entity.
        /// </summary>
        Vector3 targetPosition;

        /// <summary>
        /// Minimum speed at which the entity will move towards the target position.
        /// </summary>
        float baseCorrectiveSpeed;

        /// <summary>
        /// Amount of position error to change the goal velocity.  For example, if the entity is
        /// very far away from the target, it will move faster than if it were closer.
        /// </summary>
        float springStrength;

        /// <summary>
        /// Higher values of softness allow the constraint to be violated more.  Must be greater than zero.
        /// Sometimes, if a joint system is unstable, increasing the softness of the involved constraints will make it settle down.
        /// </summary>
        float softness;

        /// <summary>
        /// Maximum force that this constraint is capable of applying.
        /// </summary>
        float maximumForce;

        #endregion

        #region Simulation Variables
        /// <summary>
        /// The prestep-computed velocity that the motor will try to achieve.
        /// </summary>
        Vector3 goalVelocity;

        /// <summary>
        /// Maximum impulse that can be applied in a single frame.
        /// </summary>
        float maxForceDt;

        /// <summary>
        /// Maximum impulse that can be applied in a single frame, squared.
        /// This is computed in the prestep to avoid doing extra multiplies in the more-often called applyImpulse method.
        /// </summary>
        float maxForceDtSquared;

        /// <summary>
        /// Sum of forces applied to the constraint in the past.
        /// </summary>
        Vector3 accumulatedImpulse = Vector3.Zero;

        float velocityToImpulse;
        #endregion

        /// <summary>
        /// Constructs a new single body linear motor.  This motor will try to move a single entity to a goal position.
        /// </summary>
        /// <param name="entity">Entity to affect.</param>
        /// <param name="targetPosition">Initial location to move to.</param>
        /// <param name="baseCorrectiveSpeed">Minimum speed at which the entity will move towards the target position.</param>
        /// <param name="springStrength">Amount of position error to change the goal velocity.  For example, if the entity is
        /// very far away from the target, it will move faster than if it were closer.</param>
        /// <param name="softness">Higher values of softness allow the constraint to be violated more.  Must be greater than zero.
        /// Sometimes, if a joint system is unstable, increasing the softness of the involved constraints will make it settle down.</param>
        /// <param name="maximumForce">Maximum force that this constraint is capable of applying.</param>
        public SingleBodyLinearMotor(Entity entity, Vector3 targetPosition, float baseCorrectiveSpeed, float springStrength, float softness, float maximumForce)
        {
            this.myEntity = entity;
            this.targetPosition = targetPosition;
            this.baseCorrectiveSpeed = baseCorrectiveSpeed;
            this.springStrength = springStrength;
            this.softness = softness;
            this.maximumForce = maximumForce;

            collectInvolvedEntities();
            
        }


        protected override void preStep(float dt)
        {
            Vector3 separationDirection = targetPosition - entity.internalCenterPosition;
            float separationDistance = separationDirection.Length();
            if (separationDistance > Toolbox.bigEpsilon)
            {
                //The rate of correction can be based on a constant correction velocity as well as a 'spring like' correction velocity.
                //The constant correction velocity could overshoot the destination, so clamp it.
                float correctionSpeed = MathHelper.Min(baseCorrectiveSpeed, separationDistance / dt) +
                                        separationDistance * (springStrength / dt) * mySpace.simulationSettings.timeStep.timeScale;

                separationDirection /= separationDistance;

                goalVelocity = separationDirection * correctionSpeed;


            }
            else
            {
                goalVelocity = Vector3.Zero;
            }

            //Compute the maximum force that can be applied this frame.
            if (maximumForce < float.MaxValue)
            {
                maxForceDt = maximumForce * dt;
                maxForceDtSquared = maxForceDt * maxForceDt;
            }
            else
            {
                maxForceDt = float.MaxValue;
                maxForceDtSquared = float.MaxValue;
            }

            //Transforms a change in velocity to a change in momentum when multiplied.
            velocityToImpulse = 1 / (1 / myEntity.mass + softness);

            //"Warm start" the constraint by applying a first guess of the solution should be.
            myEntity.applyLinearImpulse(ref accumulatedImpulse);

        }

        protected override void applyImpulse(float dt)
        {
            //How much do we need to change the velocity to get to what we want?
            //It's easier to grasp only about the first part of this (goalVelocity - entityVelocity).
            //The softness portion allows the constraint to be violated, so the final velocity change is
            //less than what the first part suggests.
            Vector3 velocityChange = goalVelocity - myEntity.internalLinearVelocity - softness * accumulatedImpulse;

            //In terms of an impulse (an instantaneous change in momentum), what is it?
            Vector3 impulse = velocityChange * velocityToImpulse;

            Vector3 previousAccumulatedImpulse = accumulatedImpulse;
            //Sum the impulse.
            accumulatedImpulse += impulse;

            //If the impulse it takes to get to the goal is too high for the motor to handle, scale it back.
            float sumImpulseLengthSquared = accumulatedImpulse.LengthSquared();
            if (sumImpulseLengthSquared > maxForceDtSquared)
            {
                //max / impulse gives some value 0 < x < 1.  Basically, normalize the vector (divide by the length) and scale by the maximum.
                accumulatedImpulse *= maxForceDt / (float)Math.Sqrt(sumImpulseLengthSquared);

                //Since the limit was exceeded by this corrective impulse, limit it so that the accumulated impulse remains constrained.
                impulse = accumulatedImpulse - previousAccumulatedImpulse;
            }


            myEntity.applyLinearImpulse(ref impulse);

            checkForEarlyOutIterations(impulse);


        }
    }
}
